@@ -22,15 +22,15 @@ from openstack import exceptions as sdk_exc
 from osc_lib.command import command
 from osc_lib import exceptions as exc
 from osc_lib import utils
+from oslo_utils import strutils
 import six
 
 from senlinclient.common.i18n import _
-from senlinclient.common.i18n import _LI
 from senlinclient.common import utils as senlin_utils
 
 
 class ListCluster(command.Lister):
-    """List the user's clusters."""
+    """List clusters."""
 
     log = logging.getLogger(__name__ + ".ListCluster")
 
@@ -38,7 +38,7 @@ class ListCluster(command.Lister):
         parser = super(ListCluster, self).get_parser(prog_name)
         parser.add_argument(
             '--filters',
-            metavar='<key1=value1;key=value...>',
+            metavar='<"key1=value1;key2=value2...">',
             help=_("Filter parameters to apply on returned clusters. "
                    "This can be specified multiple times, or once with "
                    "parameters separated by a semicolon. The valid filter"
@@ -96,10 +96,15 @@ class ListCluster(command.Lister):
 
         clusters = senlin_client.clusters(**queries)
         formatters = {}
+        if parsed_args.global_project:
+            columns.append('project_id')
         if not parsed_args.full_id:
             formatters = {
                 'id': lambda x: x[:8]
             }
+            if 'project_id' in columns:
+                formatters['project_id'] = lambda x: x[:8]
+
         return (
             columns,
             (utils.get_item_properties(c, columns, formatters=formatters)
@@ -135,10 +140,13 @@ def _show_cluster(senlin_client, cluster_id):
         raise exc.CommandError(_('Cluster not found: %s') % cluster_id)
 
     formatters = {
+        'config': senlin_utils.json_formatter,
         'metadata': senlin_utils.json_formatter,
         'node_ids': senlin_utils.list_formatter
     }
     data = cluster.to_dict()
+    if 'is_profile_only' in data:
+        data.pop('is_profile_only')
     columns = sorted(data.keys())
     return columns, utils.get_dict_properties(data, columns,
                                               formatters=formatters)
@@ -151,6 +159,14 @@ class CreateCluster(command.ShowOne):
 
     def get_parser(self, prog_name):
         parser = super(CreateCluster, self).get_parser(prog_name)
+        parser.add_argument(
+            '--config',
+            metavar='<"key1=value1;key2=value2...">',
+            help=_('Configuration of the cluster. Default to {}. '
+                   'This can be specified multiple times, or once with '
+                   'key-value pairs separated by a semicolon.'),
+            action='append'
+        )
         parser.add_argument(
             '--min-size',
             metavar='<min-size>',
@@ -178,7 +194,7 @@ class CreateCluster(command.ShowOne):
         )
         parser.add_argument(
             '--metadata',
-            metavar='<key1=value1;key2=value2...>',
+            metavar='<"key1=value1;key2=value2...">',
             help=_('Metadata values to be attached to the cluster. '
                    'This can be specified multiple times, or once with '
                    'key-value pairs separated by a semicolon.'),
@@ -188,7 +204,7 @@ class CreateCluster(command.ShowOne):
             '--profile',
             metavar='<profile>',
             required=True,
-            help=_('Profile Id used for this cluster')
+            help=_('Default profile Id or name used for this cluster')
         )
         parser.add_argument(
             'name',
@@ -204,6 +220,7 @@ class CreateCluster(command.ShowOne):
         if parsed_args.min_size and not parsed_args.desired_capacity:
             parsed_args.desired_capacity = parsed_args.min_size
         attrs = {
+            'config': senlin_utils.format_parameters(parsed_args.config),
             'name': parsed_args.name,
             'profile_id': parsed_args.profile,
             'min_size': parsed_args.min_size,
@@ -224,11 +241,28 @@ class UpdateCluster(command.ShowOne):
 
     def get_parser(self, prog_name):
         parser = super(UpdateCluster, self).get_parser(prog_name)
+        parser.add_argument(
+            '--config',
+            metavar='<"key1=value1;key2=value2...">',
+            help=_('s of the cluster. Default to {}. '
+                   'This can be specified multiple times, or once with '
+                   'key-value pairs separated by a semicolon.'),
+            action='append'
+        )
         parser.add_argument(
             '--profile',
             metavar='<profile>',
             help=_('ID or name of new profile to use')
         )
+        parser.add_argument(
+            '--profile-only',
+            default=False, metavar='<boolean>',
+            help=_("Whether the cluster should be updated profile only. "
+                   "If false, it will be applied to all existing nodes. "
+                   "If true, any newly created nodes will use the new profile,"
+                   "but existing nodes will not be changed. Default is False.")
+
+        )
         parser.add_argument(
             '--timeout',
             metavar='<timeout>',
@@ -236,10 +270,11 @@ class UpdateCluster(command.ShowOne):
         )
         parser.add_argument(
             '--metadata',
-            metavar='<key1=value1;key2=value2...>',
-            help=_('Metadata values to be attached to the cluster. '
-                   'This can be specified multiple times, or once with '
-                   'key-value pairs separated by a semicolon'),
+            metavar='<"key1=value1;key2=value2...">',
+            help=_("Metadata values to be attached to the cluster. "
+                   "This can be specified multiple times, or once with "
+                   "key-value pairs separated by a semicolon. Use '{}' "
+                   "can clean metadata "),
             action='append'
         )
         parser.add_argument(
@@ -264,11 +299,15 @@ class UpdateCluster(command.ShowOne):
         attrs = {
             'name': parsed_args.name,
             'profile_id': parsed_args.profile,
+            'profile_only': strutils.bool_from_string(
+                parsed_args.profile_only,
+                strict=True,
+            ),
             'metadata': senlin_utils.format_parameters(parsed_args.metadata),
             'timeout': parsed_args.timeout,
         }
 
-        senlin_client.update_cluster(cluster.id, **attrs)
+        senlin_client.update_cluster(cluster, **attrs)
         return _show_cluster(senlin_client, cluster.id)
 
 
@@ -283,12 +322,17 @@ class DeleteCluster(command.Command):
             'cluster',
             metavar='<cluster>',
             nargs='+',
-            help=_('Name or ID of cluster(s) to delete')
+            help=_('Name or ID of cluster(s) to delete.')
+        )
+        parser.add_argument(
+            '--force-delete',
+            action='store_true',
+            help=_('Force to delete cluster(s).')
         )
         parser.add_argument(
             '--force',
             action='store_true',
-            help=_('Skip yes/no prompt (assume yes)')
+            help=_('Skip yes/no prompt (assume yes).')
         )
         return parser
 
@@ -305,16 +349,17 @@ class DeleteCluster(command.Command):
                 if not prompt_response.startswith('y'):
                     return
         except KeyboardInterrupt:  # Ctrl-c
-            self.log.info(_LI('Ctrl-c detected.'))
+            self.log.info('Ctrl-c detected.')
             return
         except EOFError:  # Ctrl-d
-            self.log.info(_LI('Ctrl-d detected'))
+            self.log.info('Ctrl-d detected')
             return
 
         result = {}
         for cid in parsed_args.cluster:
             try:
-                cluster = senlin_client.delete_cluster(cid, False)
+                cluster = senlin_client.delete_cluster(
+                    cid, False, parsed_args.force_delete)
                 result[cid] = ('OK', cluster.location.split('/')[-1])
             except Exception as ex:
                 result[cid] = ('ERROR', six.text_type(ex))
@@ -399,6 +444,12 @@ class ResizeCluster(command.Command):
         max_size = parsed_args.max_size
         min_step = parsed_args.min_step
 
+        if sum(v is not None for v in (capacity, adjustment, percentage,
+                                       min_size, max_size)) == 0:
+            raise exc.CommandError(_("At least one parameter of 'capacity', "
+                                     "'adjustment', 'percentage', 'min_size' "
+                                     "and 'max_size' should be specified."))
+
         if sum(v is not None for v in (capacity, adjustment, percentage)) > 1:
             raise exc.CommandError(_("Only one of 'capacity', 'adjustment' and"
                                      " 'percentage' can be specified."))
@@ -425,10 +476,9 @@ class ResizeCluster(command.Command):
             action_args['adjustment_type'] = 'CHANGE_IN_PERCENTAGE'
             action_args['number'] = percentage
 
-        if min_step is not None:
-            if percentage is None:
-                raise exc.CommandError(_('Min step is only used with '
-                                         'percentage.'))
+        if min_step is not None and percentage is None:
+            raise exc.CommandError(_('Min step is only used with '
+                                     'percentage.'))
 
         if min_size is not None:
             if min_size < 0:
@@ -454,7 +504,10 @@ class ResizeCluster(command.Command):
         action_args['strict'] = parsed_args.strict
 
         resp = senlin_client.cluster_resize(parsed_args.cluster, **action_args)
-        print('Request accepted by action: %s' % resp['action'])
+        if 'action' in resp:
+            print('Request accepted by action: %s' % resp['action'])
+        else:
+            print('Request error: %s' % resp)
 
 
 class ScaleInCluster(command.Command):
@@ -482,7 +535,10 @@ class ScaleInCluster(command.Command):
 
         resp = senlin_client.cluster_scale_in(parsed_args.cluster,
                                               parsed_args.count)
-        print('Request accepted by action %s' % resp['action'])
+        if 'action' in resp:
+            print('Request accepted by action: %s' % resp['action'])
+        else:
+            print('Request error: %s' % resp)
 
 
 class ScaleOutCluster(command.Command):
@@ -510,7 +566,10 @@ class ScaleOutCluster(command.Command):
 
         resp = senlin_client.cluster_scale_out(parsed_args.cluster,
                                                parsed_args.count)
-        print('Request accepted by action %s' % resp['action'])
+        if 'action' in resp:
+            print('Request accepted by action: %s' % resp['action'])
+        else:
+            print('Request error: %s' % resp)
 
 
 class ClusterPolicyAttach(command.Command):
@@ -522,8 +581,8 @@ class ClusterPolicyAttach(command.Command):
         parser = super(ClusterPolicyAttach, self).get_parser(prog_name)
         parser.add_argument(
             '--enabled',
+            metavar='<boolean>',
             default=True,
-            action="store_true",
             help=_('Whether the policy should be enabled once attached. '
                    'Default to True')
         )
@@ -545,13 +604,17 @@ class ClusterPolicyAttach(command.Command):
         senlin_client = self.app.client_manager.clustering
 
         kwargs = {
-            'enabled': parsed_args.enabled,
+            'enabled': strutils.bool_from_string(parsed_args.enabled,
+                                                 strict=True),
         }
 
         resp = senlin_client.cluster_attach_policy(parsed_args.cluster,
                                                    parsed_args.policy,
                                                    **kwargs)
-        print('Request accepted by action: %s' % resp['action'])
+        if 'action' in resp:
+            print('Request accepted by action: %s' % resp['action'])
+        else:
+            print('Request error: %s' % resp)
 
 
 class ClusterPolicyDetach(command.Command):
@@ -579,7 +642,10 @@ class ClusterPolicyDetach(command.Command):
         senlin_client = self.app.client_manager.clustering
         resp = senlin_client.cluster_detach_policy(parsed_args.cluster,
                                                    parsed_args.policy)
-        print('Request accepted by action: %s' % resp['action'])
+        if 'action' in resp:
+            print('Request accepted by action: %s' % resp['action'])
+        else:
+            print('Request error: %s' % resp)
 
 
 class ClusterNodeList(command.Lister):
@@ -591,18 +657,21 @@ class ClusterNodeList(command.Lister):
         parser = super(ClusterNodeList, self).get_parser(prog_name)
         parser.add_argument(
             '--filters',
-            metavar='<key1=value;key2=value2...>',
-            help=_('Filter parameters to apply on returned nodes. '
-                   'This can be specified multiple times, or once with '
-                   'parameters separated by a semicolon'),
+            metavar='<key1=value1;key2=value2...>',
+            help=_("Filter parameters to apply on returned nodes. "
+                   "This can be specified multiple times, or once with "
+                   "parameters separated by a semicolon. The valid filter "
+                   "keys are: ['status', 'name']"),
             action='append'
         )
         parser.add_argument(
             '--sort',
             metavar='<key>[:<direction>]',
-            help=_('Sorting option which is a string containing a list of '
-                   'keys separated by commas. Each key can be optionally '
-                   'appended by a sort direction (:asc or :desc)')
+            help=_("Sorting option which is a string containing a list of "
+                   "keys separated by commas. Each key can be optionally "
+                   "appended by a sort direction (:asc or :desc)' The valid "
+                   "sort keys are:['index', 'name', 'status', 'init_at', "
+                   "'created_at', 'updated_at']")
         )
         parser.add_argument(
             '--limit',
@@ -682,7 +751,10 @@ class ClusterNodeAdd(command.Command):
         senlin_client = self.app.client_manager.clustering
         node_ids = parsed_args.nodes.split(',')
         resp = senlin_client.cluster_add_nodes(parsed_args.cluster, node_ids)
-        print('Request accepted by action: %s' % resp['action'])
+        if 'action' in resp:
+            print('Request accepted by action: %s' % resp['action'])
+        else:
+            print('Request error: %s' % resp)
 
 
 class ClusterNodeDel(command.Command):
@@ -698,6 +770,14 @@ class ClusterNodeDel(command.Command):
             help=_('Name or ID of nodes to be deleted; multiple nodes can be '
                    'separated with ","')
         )
+        parser.add_argument(
+            '-d',
+            '--destroy-after-deletion',
+            required=False,
+            default=False,
+            help=_('Whether nodes should be destroyed after deleted. '
+                   'Default is False.')
+        )
         parser.add_argument(
             'cluster',
             metavar='<cluster>',
@@ -709,8 +789,54 @@ class ClusterNodeDel(command.Command):
         self.log.debug("take_action(%s)", parsed_args)
         senlin_client = self.app.client_manager.clustering
         node_ids = parsed_args.nodes.split(',')
-        resp = senlin_client.cluster_del_nodes(parsed_args.cluster, node_ids)
-        print('Request accepted by action: %s' % resp['action'])
+        destroy = parsed_args.destroy_after_deletion
+        destroy = strutils.bool_from_string(destroy, strict=True)
+        kwargs = {"destroy_after_deletion": destroy}
+        resp = senlin_client.cluster_del_nodes(parsed_args.cluster, node_ids,
+                                               **kwargs)
+        if 'action' in resp:
+            print('Request accepted by action: %s' % resp['action'])
+        else:
+            print('Request error: %s' % resp)
+
+
+class ClusterNodeReplace(command.Command):
+    """Replace the nodes in a cluster with specified nodes."""
+    log = logging.getLogger(__name__ + ".ClusterNodeReplace")
+
+    def get_parser(self, prog_name):
+        parser = super(ClusterNodeReplace, self).get_parser(prog_name)
+        parser.add_argument(
+            '--nodes',
+            metavar='<OLD_NODE1=NEW_NODE1>',
+            required=True,
+            help=_("OLD_NODE is the name or ID of a node to be replaced, "
+                   "NEW_NODE is the name or ID of a node as replacement. "
+                   "This can be specified multiple times, or once with "
+                   "node-pairs separated by a comma ','."),
+            action='append'
+        )
+        parser.add_argument(
+            'cluster',
+            metavar='<cluster>',
+            help=_('Name or ID of cluster to operate on')
+        )
+        return parser
+
+    def take_action(self, parsed_args):
+        self.log.debug("take_action(%s)", parsed_args)
+        senlin_client = self.app.client_manager.clustering
+        nodepairs = {}
+        for nodepair in parsed_args.nodes:
+            key = nodepair.split('=')[0]
+            value = nodepair.split('=')[1]
+            nodepairs[key] = value
+        resp = senlin_client.cluster_replace_nodes(parsed_args.cluster,
+                                                   nodepairs)
+        if 'action' in resp:
+            print('Request accepted by action: %s' % resp['action'])
+        else:
+            print('Request error: %s' % resp)
 
 
 class CheckCluster(command.Command):
@@ -735,9 +861,12 @@ class CheckCluster(command.Command):
                 resp = senlin_client.check_cluster(cid)
             except sdk_exc.ResourceNotFound:
                 raise exc.CommandError(_('Cluster not found: %s') % cid)
-            print('Cluster check request on cluster %(cid)s is accepted by '
-                  'action %(action)s.'
-                  % {'cid': cid, 'action': resp['action']})
+            if 'action' in resp:
+                print('Cluster check request on cluster %(cid)s is '
+                      'accepted by action %(action)s.'
+                      % {'cid': cid, 'action': resp['action']})
+            else:
+                print('Request error: %s' % resp)
 
 
 class RecoverCluster(command.Command):
@@ -752,23 +881,40 @@ class RecoverCluster(command.Command):
             nargs='+',
             help=_('ID or name of cluster(s) to operate on.')
         )
+
+        parser.add_argument(
+            '--check',
+            metavar='<boolean>',
+            default=False,
+            help=_("Whether the cluster should check it's nodes status before "
+                   "doing cluster recover. Default is false")
+        )
+
         return parser
 
     def take_action(self, parsed_args):
         self.log.debug("take_action(%s)", parsed_args)
         senlin_client = self.app.client_manager.clustering
+
+        params = {
+            'check': strutils.bool_from_string(parsed_args.check, strict=True)
+        }
+
         for cid in parsed_args.cluster:
             try:
-                resp = senlin_client.recover_cluster(cid)
+                resp = senlin_client.recover_cluster(cid, **params)
             except sdk_exc.ResourceNotFound:
                 raise exc.CommandError(_('Cluster not found: %s') % cid)
-            print('Cluster recover request on cluster %(cid)s is accepted by '
-                  'action %(action)s.'
-                  % {'cid': cid, 'action': resp['action']})
+            if 'action' in resp:
+                print('Cluster recover request on cluster %(cid)s is '
+                      'accepted by action %(action)s.'
+                      % {'cid': cid, 'action': resp['action']})
+            else:
+                print('Request error: %s' % resp)
 
 
 class ClusterCollect(command.Lister):
-    """Recover the cluster(s)."""
+    """Collect attributes across a cluster."""
     log = logging.getLogger(__name__ + ".ClusterCollect")
 
     def get_parser(self, prog_name):
@@ -808,6 +954,53 @@ class ClusterCollect(command.Lister):
                  for a in attrs))
 
 
+class ClusterOp(command.Lister):
+    """Perform an operation on all nodes across a cluster."""
+    log = logging.getLogger(__name__ + ".ClusterOp")
+
+    def get_parser(self, prog_name):
+        parser = super(ClusterOp, self).get_parser(prog_name)
+        parser.add_argument(
+            '--operation',
+            metavar='<operation>',
+            required=True,
+            help=_('Operation to be performed on the cluster')
+        )
+        parser.add_argument(
+            '--params',
+            metavar='<key1=value1;key2=value2...>',
+            help=_("Parameters to for the specified operation. "
+                   "This can be specified multiple times, or once with "
+                   "parameters separated by a semicolon."),
+            action='append'
+        )
+        parser.add_argument(
+            'cluster',
+            metavar='<cluster>',
+            help=_('ID or name of cluster to operate on.')
+        )
+        return parser
+
+    def take_action(self, parsed_args):
+        self.log.debug("take_action(%s)", parsed_args)
+        senlin_client = self.app.client_manager.clustering
+        cid = parsed_args.cluster
+        if parsed_args.params:
+            params = senlin_utils.format_parameters(parsed_args.params)
+        else:
+            params = {}
+
+        try:
+            resp = senlin_client.perform_operation_on_cluster(
+                cid, parsed_args.operation, **params)
+        except sdk_exc.ResourceNotFound:
+            raise exc.CommandError(_('Cluster not found: %s') % cid)
+        if 'action' in resp:
+            print('Request accepted by action: %s' % resp['action'])
+        else:
+            print('Request error: %s' % resp)
+
+
 class ClusterRun(command.Command):
     """Run scripts on cluster."""
     log = logging.getLogger(__name__ + ".ClusterRun")
@@ -886,12 +1079,11 @@ class ClusterRun(command.Command):
         except sdk_exc.ResourceNotFound:
             raise exc.CommandError(_("Cluster not found: %s") % args.cluster)
 
-        script = None
         try:
             f = open(args.script, 'r')
             script = f.read()
         except Exception:
-            raise exc.CommandError(_("Cound not open script file: %s") %
+            raise exc.CommandError(_("Could not open script file: %s") %
                                    args.script)
 
         tasks = dict()
@@ -950,7 +1142,7 @@ class ClusterRun(command.Command):
                                     "network.") % {'node': node_id}
                 return
             else:
-                addresses = list(six.itervalues(addr))[0]
+                addresses = list(addr.values())[0]
 
         # Select the address in the selected network.
         # If the extension is not present, we assume the address to be
